﻿using Amqp;
using Amqp.Framing;
using Amqp.Sasl;
using AmqpTest.Configs;
using Microsoft.Extensions.Logging;
using Microsoft.Extensions.Options;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Security.Cryptography;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

namespace AmqpTest.Amqp
{
	/// <summary>
	/// 用于处理阿里云平台amqp消息队列的数据（负责接收）
	/// </summary>
	public class AmqpSubscribe
	{
		private readonly ILogger<AmqpSubscribe> _logger;
		private readonly IotConfig _configIot;

		public AmqpSubscribe(ILogger<AmqpSubscribe> logger, IOptions<IotConfig> optConfigIot)
		{
			_logger = logger;
			_configIot = optConfigIot.Value;
		}

		/// <summary>
		/// 开始接收iot平台amqp推送的数据
		/// </summary>
		/// <returns></returns>
		public async Task BeginListen()
		{
			try
			{
				long timestamp = GetCurrentMilliseconds();
				string param = "authId=" + _configIot.AccessKey + "&timestamp=" + timestamp;
				string password = DoSign(param, _configIot.AccessSecret, "HmacMD5");
				string iotInstanceId = !string.IsNullOrEmpty(_configIot.IotInstanceId) ? "iotInstanceId=" + _configIot.IotInstanceId + "," : "";

				string username = _configIot.ClientId + "|" + iotInstanceId + "authMode=aksign,signMethod=hmacmd5,consumerGroupId=" + _configIot.ConsumerGroupId
					+ ",authId=" + _configIot.AccessKey + ",timestamp=" + timestamp + "|";

				await DoConnectAmqpAsync(username, password);
			}
			catch (Exception ex)
			{
				_logger.LogError($"BeginListen 处理AMQP数据发生异常 {ex.Message}\n{ex.StackTrace}");
			}
		}

		/// <summary>
		/// 获取当前时间的时间戳
		/// </summary>
		/// <returns></returns>
		private long GetCurrentMilliseconds()
		{
			DateTime dt1970 = new DateTime(1970, 1, 1);
			DateTime current = DateTime.Now;
			return (long)(current - dt1970).TotalMilliseconds;
		}

		public string DoSign(string param, string accessSecret, string signMethod)
		{
			byte[] key = Encoding.UTF8.GetBytes(accessSecret);
			byte[] signContent = Encoding.UTF8.GetBytes(param);
			var hmac = new HMACMD5(key);
			byte[] hashBytes = hmac.ComputeHash(signContent);
			return Convert.ToBase64String(hashBytes);
		}

		private async Task DoConnectAmqpAsync(string username, string password)
		{
			string host = $"{_configIot.Uid}.iot-amqp.{_configIot.RegionId}.aliyuncs.com";
			int port = Convert.ToInt32(_configIot.Port);

			var address = new Address(host, port, username, password);
			//Create Connection
			ConnectionFactory cf = new ConnectionFactory();
			//use local tls if neccessary
			//cf.SSL.ClientCertificates.Add(GetCert());
			//cf.SSL.RemoteCertificateValidationCallback = ValidateServerCertificate;
			cf.SASL.Profile = SaslProfile.External;
			cf.AMQP.IdleTimeout = 120000;
			cf.AMQP.ContainerId = "client.1.2";
			cf.AMQP.HostName = "contoso.com";
			cf.AMQP.MaxFrameSize = 8 * 1024;

			var connection = await cf.CreateAsync(address);
			if (connection.ConnectionState.Equals(ConnectionState.OpenSent))
			{
				_logger.LogInformation("Open frame was received.");
			}
			//Connection Exception Closed
			connection.AddClosedCallback(ConnClosed);

			//Receive Message
			await DoReceiveAsync(connection);
		}

		private async Task DoReceiveAsync(Connection connection)
		{
			var session = new Session(connection);
			var receiver = new ReceiverLink(session, "queueName", null);
			_logger.LogInformation("开始接收设备信息!");
			int maxCount = Convert.ToInt32(_configIot.MaxDegreeOfParallelism);
			//LimitedConcurrencyLevelTaskScheduler lcts = new LimitedConcurrencyLevelTaskScheduler(maxCount);
			//TaskFactory taskFactory = new TaskFactory(lcts);

			try
			{
				int count = 0;
				var sw = new Stopwatch();
				sw.Start();

				//只是控制该线程不退出
				while (true)
				{
					receiver.Start(100, (link, message) =>
					{
						var messageId = message.ApplicationProperties["messageId"].ToString();
						var topic = message.ApplicationProperties["topic"].ToString();
						var body = Encoding.UTF8.GetString((byte[])message.Body);

						link.Accept(message);

						_logger.LogInformation($"message arrived, topic= {topic}, messageId= {messageId}, body= {body}");
						count++;
					});

					if (sw.ElapsedMilliseconds > 60000)
					{
						_logger.LogWarning($"约1分钟处理 {count} 个请求");
						sw.Restart();
						count = 0;
					}

					await Task.Delay(100).ConfigureAwait(false); //睡眠值用于控制计时的准确性，和iot消息处理效率无关
				}
			}
			catch (Exception)
			{
				throw;
			}
			finally
			{
				connection?.Close();
			}
		}

		private void ConnClosed(IAmqpObject _, Error e)
		{
			_logger.LogError($"连接关闭:{e?.ToString()}");
		}
	}
}
